#!/usr/bin/env python
import os
import sys

from django.conf import settings
from django.core.management import execute_from_command_line


def main():
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "yawn.settings.debug")

    # check if yawn is in installed apps, and bail if it is not
    if 'yawn' not in settings.INSTALLED_APPS:
        print("Please check your DJANGO_SETTINGS_MODULE environment variable.\n"
              "Make sure 'yawn' is in your INSTALLED_APPS.\n"
              "Generally, your settings file should start with 'from yawn.settings.base import *'")
        sys.exit(1)

    print('YAWN workflow management tool')
    if os.environ['DJANGO_SETTINGS_MODULE'] == 'yawn.settings.debug':
        print('  Running in DEBUG mode')

    # run the django manage.py command line
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



import datetime
import pytest

from .. import cron


@pytest.mark.parametrize('expression,expected', [
    ('*', list(range(7))),
    ('1-3', [1, 2, 3]),
    ('*/2', [0, 2, 4, 6]),
    ('2-5/2', [2, 4]),
    ('1,*/6', [0, 1, 6]),
])
def test_parse_valid(expression, expected):
    assert cron.parse(expression, 7) == expected


@pytest.mark.parametrize('expression', [
    '*-1', '1-', '-1',  # invalid range
    '2a',  # invalid integer
    '7',  # out of range
    '/2', '2/',  # empty numerator/denominator
    '2,',  # empty part
    '5/7,',  # no values
])
def test_parse_invalid(expression):
    with pytest.raises(ValueError):
        cron.parse(expression, 7)


@pytest.mark.parametrize('current_time,next_run', [
    # there is a run this hour
    (datetime.datetime(2016, 11, 28, 4, 15), datetime.datetime(2016, 11, 28, 4, 30)),
    # there is a run today
    (datetime.datetime(2016, 11, 28, 4, 59), datetime.datetime(2016, 11, 28, 5, 0)),
    # there is a run on another day
    (datetime.datetime(2016, 11, 28, 5, 59), datetime.datetime(2016, 11, 29, 2, 0)),
])
def test_cron_next_run(current_time, next_run):
    tab = cron.Crontab('*/15 2-5 1,2')
    assert tab.next_run(current_time) == next_run



"""
Workflows:
A, B, C succeed, D is waiting -> Running
A, B, C succeed, D is running -> Running
A, B, C failed, D is queued -> Running
A, B, C, D succeed -> Succeeded
A, B failed, C, D upstream_failed -> Failed
"""
import datetime
from unittest import mock

import pytz
from django.utils import timezone

from yawn.task.models import Template, Task
from yawn.workflow.models import Workflow, WorkflowName, Run


def test_first_ready():
    name = WorkflowName.objects.create(name='workflow1')

    # not active, shouldn't get selected
    Workflow.objects.create(
        name=name, version=1, schedule_active=False,
        next_run=timezone.now() - datetime.timedelta(hours=1)
    )
    # future run date
    Workflow.objects.create(
        name=name, version=2, schedule_active=True,
        next_run=timezone.now() + datetime.timedelta(hours=1)
    )
    ready_workflow = Workflow.objects.create(
        name=name, version=3, schedule_active=True,
        next_run=timezone.now() - datetime.timedelta(seconds=1)
    )

    workflow = Workflow.first_ready()
    assert workflow.id == ready_workflow.id


@mock.patch('yawn.workflow.models.Crontab')
def test_next_run(mock_cron):
    next_run = timezone.datetime(2011, 1, 1, tzinfo=pytz.UTC)
    mock_cron().next_run.return_value = next_run

    name = WorkflowName.objects.create(name='workflow1')
    workflow = Workflow.objects.create(
        name=name, version=2, schedule_active=True,
        schedule='*'
    )
    assert workflow.next_run == next_run


def test_new_version():
    name = WorkflowName.objects.create(name='workflow1')
    workflow_1 = name.new_version(schedule='0 0 *', schedule_active=True)
    assert workflow_1.next_run
    assert workflow_1.version == 1
    workflow_2 = name.new_version()
    assert workflow_2.version == 2
    workflow_1.refresh_from_db()
    assert workflow_1.schedule_active is False


def test_submit_run():
    # create a workflow
    name = WorkflowName.objects.create(name='workflow1')
    workflow = name.new_version(parameters={'parent': True, 'child': False})
    task1 = Template.objects.create(workflow=workflow, name='task1', command=[''])
    task2 = Template.objects.create(workflow=workflow, name='task2', command=[''])
    task2.upstream.add(task1)

    # submit a run
    run = workflow.submit_run(parameters={'child': True})
    tasks = list(run.task_set.all())

    # check that tasks are copied to the run correctly
    assert run.parameters == {'parent': True, 'child': True}
    assert len(tasks) == 2
    assert tasks[0].template == task1
    assert tasks[0].status == Task.QUEUED
    assert tasks[1].template == task2
    assert tasks[1].status == Task.WAITING


def test_statuses():
    # create a workflow
    name = WorkflowName.objects.create(name='workflow1')
    workflow = name.new_version(parameters={'parent': True, 'child': False})
    task1 = Template.objects.create(workflow=workflow, name='task1', command=[''])
    task2 = Template.objects.create(workflow=workflow, name='task2', command=[''])
    task2.upstream.add(task1)

    run = workflow.submit_run(parameters={'child': True})
    task1, task2 = list(run.task_set.all())

    run.update_status()
    assert run.status == Run.RUNNING

    task1.status = Task.FAILED
    task1.save()
    run.update_status()
    assert run.status == Run.RUNNING

    task2.status = Task.UPSTREAM_FAILED
    task2.save()
    run.update_status()
    assert run.status == Run.FAILED

    task1.status = Task.SUCCEEDED
    task1.save()
    task2.status = Task.SUCCEEDED
    task2.save()
    run.update_status()
    assert run.status == Run.SUCCEEDED



from django.db import models
from django.contrib.postgres import fields
from django.db.models import functions

from yawn.utilities import cron, database
from yawn.utilities.cron import Crontab


class WorkflowName(models.Model):
    name = models.TextField(unique=True)
    current_version = models.OneToOneField('Workflow', on_delete=models.CASCADE,
                                           null=True, related_name='is_current')

    def new_version(self, **kwargs):
        """Create a new version of a workflow"""
        version = 0
        if self.current_version_id:
            version = self.current_version.version

            # disable the past schedule
            self.current_version.schedule_active = False
            self.current_version.save()

        workflow = Workflow.objects.create(name=self, version=version + 1, **kwargs)
        self.current_version = workflow
        self.save()
        return workflow


class Workflow(models.Model):
    class Meta:
        unique_together = (('name', 'version'),)

    name = models.ForeignKey(WorkflowName, models.PROTECT)
    version = models.IntegerField(editable=False)  # serializer read-only

    # scheduling is completely optional:
    schedule_active = models.BooleanField(default=False)
    schedule = models.TextField(null=True, validators=[cron.cron_validator])
    next_run = models.DateTimeField(null=True)

    # parameters
    parameters = fields.JSONField(default=dict)

    def save(self, **kwargs):
        if self.schedule_active:
            if not self.next_run:
                self.next_run = Crontab(self.schedule).next_run(database.current_time())
        else:
            self.next_run = None
        super().save(**kwargs)

    @classmethod
    def first_ready(cls):
        ready_templates = list(cls.objects.raw("""
            SELECT * FROM yawn_workflow
            WHERE schedule_active
            AND next_run < transaction_timestamp()
            ORDER BY next_run LIMIT 1
            FOR UPDATE SKIP LOCKED -- requires PG 9.5+
        """))
        return ready_templates[0] if len(ready_templates) else None

    def submit_run(self, parameters=None, scheduled_time=None):
        """Create a run of this template"""
        from yawn.task.models import Task

        run_parameters = self.parameters.copy()
        run_parameters.update(parameters or {})

        run = Run.objects.create(
            workflow=self,
            submitted_time=functions.Now(),
            scheduled_time=scheduled_time,
            parameters=run_parameters,
        )
        for template in self.template_set.all():
            task = Task.objects.create(
                run=run,
                template=template,
            )
            if not template.upstream.exists():
                task.enqueue()

        # refresh to get the actual DB submitted time
        run.refresh_from_db()
        return run

    def __str__(self):
        return '{} v{}'.format(self.name.name, self.version)


class Run(models.Model):
    RUNNING = 'running'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    STATUS_CHOICES = [(x, x) for x in (RUNNING, SUCCEEDED, FAILED)]

    workflow = models.ForeignKey(Workflow, models.PROTECT)
    submitted_time = models.DateTimeField()
    scheduled_time = models.DateTimeField(null=True)
    status = models.TextField(default=RUNNING, choices=STATUS_CHOICES)
    parameters = fields.JSONField(default=dict)

    def update_status(self):
        from yawn.task.models import Task

        task_statuses = self.task_set.all().values_list('status', flat=True)

        if all([status == Task.SUCCEEDED for status in task_statuses]):
            self.status = self.SUCCEEDED
        elif any([status in [Task.WAITING, Task.QUEUED, Task.RUNNING] for status in task_statuses]):
            self.status = self.RUNNING
        else:
            self.status = self.FAILED

        self.save()

    def __str__(self):
        return 'Workflow {} Run {}'.format(self.workflow_id, self.id)



import datetime

import pytest
from django.db import connection

from yawn.utilities import database
from yawn.utilities.database import close_on_exception


@pytest.mark.no_transaction
def test_close_on_exception():

    @close_on_exception
    def example_disconnect():
        with connection.cursor() as cursor:

            # kill the current connection, which will raise a django.db.OperationalError
            cursor.execute('select pg_terminate_backend(pg_backend_pid())')

            # which means we never get here
            cursor.execute('select 1')
            assert False, 'exception was not raised'

    # this will cause a disconnect:
    example_disconnect()

    # but the exception is caught, and on retry the database reconnects:
    with connection.cursor() as cursor:
        cursor.execute('select 1')


def test_current_time():
    assert isinstance(database.current_time(), datetime.datetime)



import os
import socket

import logging
from django.core.management.base import BaseCommand

from yawn.worker.main import Main


class Command(BaseCommand):
    help = 'Run the worker to execute tasks'

    def add_arguments(self, parser):
        name_default = '{} {}'.format(socket.gethostname(), os.getpid())
        parser.add_argument('--concurrency', type=int, help='Number of concurrent tasks to run', default=1)
        parser.add_argument('--name', type=str, help='Worker name to register', default=name_default)
        parser.add_argument('--queues', type=str, nargs='+', default=['default'], metavar='QUEUE',
                            help='Space-separated list of queues to consume from')

    def handle(self, *args, **options):
        log_level = logging.INFO
        if options['verbosity'] == 0:
            log_level = logging.ERROR
        if options['verbosity'] > 1:
            log_level = logging.DEBUG
        logging.basicConfig(level=log_level, format='%(message)s')

        worker = Main(
            options['concurrency'],
            options['name'],
            options['queues'],
        )
        worker.run()



import pytest

from yawn.task.models import Template
from yawn.worker.models import Worker, Queue, Message
from yawn.workflow.models import WorkflowName


def test_get_workers(client):
    worker = Worker.objects.create(name='worker1')
    response = client.get('/api/workers/')
    assert response.status_code == 200
    workers = response.data['results']
    assert len(workers) == 1
    assert workers[0]['name'] == worker.name
    assert workers[0]['status'] == worker.status


@pytest.fixture()
def queue():
    queue = Queue.objects.create(name='queue1')

    # need all this for a task...
    name = WorkflowName.objects.create(name='workflow1')
    workflow = name.new_version()
    Template.objects.create(workflow=workflow, name='task1', command=[''])
    run = workflow.submit_run()
    task = run.task_set.first()

    # create some messages
    Message.objects.create(task=task, queue=queue)
    Message.objects.create(task=task, queue=queue)
    Message.objects.create(task=task, queue=Queue.get_default_queue())
    return queue


def test_get_queues(client, queue):
    response = client.get('/api/queues/')
    assert response.status_code == 200
    queues = response.data['results']
    assert len(queues) == 2
    assert queues[1]['name'] == queue.name
    assert queues[1]['message_count'] == 2



import enum
import signal
import typing  # NOQA
import collections

import time
from django.db import transaction
from django.db.models import functions

from yawn.task.models import Task, Execution
from yawn.worker.models import Worker, Queue
from yawn.worker.executor import Manager, Result  # NOQA
from yawn.workflow.models import Workflow
from yawn.utilities import logger
from yawn.utilities.cron import Crontab
from yawn.utilities.database import close_on_exception


class State(enum.Enum):
    stopped = 'Stopped'
    running = 'Running'
    shutdown = 'Shutting down'
    terminate = 'Terminating'


class Main:
    results = None  # type: typing.Deque[Result]
    executor = None
    state = State.stopped
    worker = None
    timeout = 60  # seconds

    def __init__(self, concurrency, name, queues: list):
        self.concurrency = concurrency
        self.results = collections.deque()
        self.name = name
        self.queue_ids = []
        for name in queues:
            queue, created = Queue.objects.get_or_create(name=name)
            self.queue_ids.append(queue.id)

    def run(self):
        """
        Main event loop.

        Why not threads/gevent/asyncio?
            Because this is simple and maybe more efficient
        """
        self.handle_signals()
        self.executor = Manager()
        self.state = State.running
        loop_start = time.time()
        logger.warning('Starting YAWN worker with concurrency=%s', self.concurrency)

        while True:

            if self.state == State.running:

                # update own status and check for lost workers
                self.update_worker()

                # check for tasks that should be queued
                self.schedule_workflows()

                # run queued tasks
                self.start_tasks()

                self.mark_terminated()

            elif not self.executor.get_running_ids():
                # shutdown: exit loop once all tasks are finished
                break

            # check status on running tasks
            self.results.extend(self.executor.read_output())
            self.save_results()

            # don't run the loop more than once per second
            loop_duration = time.time() - loop_start
            time.sleep(max(1 - loop_duration, 0))
            loop_start += loop_duration

        logger.warning('Exiting')
        self.worker.status = Worker.EXITED
        self.worker.save()

    def handle_signals(self):
        """
        Handle signals:
        - SIGINT waits for running tasks to stop, then exits
        - SIGTERM kills running tasks, saves to the database, and exits
        """

        def handle_sigterm(*args):
            self.state = State.terminate
            # kill running tasks
            self.executor.mark_terminated(self.executor.get_running_ids())
            logger.warning('Received SIGTERM, killing running tasks and exiting.')

        def handle_sigint(*args):
            self.state = State.shutdown
            logger.warning('Received SIGINT, shutting down after all tasks have finished.')
            logger.warning('Press CTL-C again to shut down immediately.')
            signal.signal(signal.SIGINT, handle_sigterm)

        signal.signal(signal.SIGINT, handle_sigint)
        signal.signal(signal.SIGTERM, handle_sigterm)

    @close_on_exception
    def schedule_workflows(self):
        """
        Check for workflow templates with a crontab that are ready to run.
        - Query and lock the ready template.
        - Create a run and associated tasks from the template.
        - Update the template's next_run attribute.
        """
        with transaction.atomic():
            workflow = Workflow.first_ready()
            if not workflow:
                return  # there is no workflow ready to run

            run = workflow.submit_run(scheduled_time=workflow.next_run)
            workflow.next_run = Crontab(workflow.schedule).next_run(run.submitted_time)
            workflow.save()

    @close_on_exception
    def start_tasks(self):
        """
        Check for queued tasks that we can run here and start them.
        """
        if len(self.executor.get_running_ids()) >= self.concurrency:
            return  # this worker is already running enough tasks

        with transaction.atomic():

            task = Task.first_queued(self.queue_ids)
            if not task:
                return  # there are no tasks ready to run

            execution = task.start_execution(self.worker)

        logger.info('Starting task %s execution %s', task.id, execution.id)
        self.executor.start_subprocess(
            execution_id=execution.id,
            command=execution.task.template.command,
            environment=execution.task.run.parameters if execution.task.run else {},
            timeout=execution.task.template.timeout
        )

    @close_on_exception
    def update_worker(self):
        """
        Look for executors where the connection has broken and tasks need to be re-submitted.
        """
        if not self.worker:
            self.worker = Worker.objects.create(
                name=self.name,
                start_timestamp=functions.Now(),
                last_heartbeat=functions.Now()
            )
        else:
            self.worker.refresh_from_db()

        if self.worker.status == Worker.LOST:
            # someone else marked us as lost, terminate all tasks and exit
            logger.warning('Marked as lost, committing harakiri')
            self.state = State.terminate
            self.executor.mark_terminated(self.executor.get_running_ids())
            return

        # update our timestamp so no one marks us as lost
        self.worker.last_heartbeat = functions.Now()
        self.worker.save()

        # look for lost workers and re-queue their tasks
        Worker.find_lost(self.timeout)

    @close_on_exception
    def save_results(self):
        """Persist the subprocess results to the database"""
        for _ in range(len(self.results)):
            with transaction.atomic():
                result = self.results[0]
                if result.stdout or result.stderr:
                    Execution.update_output(result.execution_id, result.stdout, result.stderr)
                if result.returncode is not None:
                    execution = Execution.objects.get(id=result.execution_id)
                    logger.info('Task %s execution %s exited with code %s',
                                execution.task.id, execution.id, result.returncode)
                    execution.mark_finished(result.returncode)
                # pop after doing the updates have completed successfully
                # in the case of an exception before here, the updates will be retried
                self.results.popleft()

    @close_on_exception
    def mark_terminated(self):
        """Kill any tasks marked for termination by the user"""
        killed_execution_ids = Execution.objects.filter(
            id__in=self.executor.get_running_ids(),
            status=Execution.KILLED,
        ).values_list('id', flat=True)
        if killed_execution_ids:
            self.executor.mark_terminated(killed_execution_ids)



from rest_framework import viewsets

from yawn.task.models import Task, Execution
from yawn.task.serializers import TaskDetailSerializer, ExecutionListSerializer


class TaskViewSet(viewsets.GenericViewSet,
                  viewsets.mixins.RetrieveModelMixin,
                  viewsets.mixins.UpdateModelMixin):
    """
    GET a task, and its executions. PATCH to re-run or terminate.
    """
    queryset = Task.objects.all().order_by('id').prefetch_related(
        'execution_set__worker').select_related('template__workflow__name')

    serializer_class = TaskDetailSerializer


class ExecutionViewSet(viewsets.GenericViewSet,
                       viewsets.mixins.ListModelMixin):
    """
    GET a list of Executions
    """
    queryset = Execution.objects.all().order_by('-id').select_related(
        'worker', 'task__template__workflow__name')

    serializer_class = ExecutionListSerializer

    def get_queryset(self):
        """
        Optionally filter to the executions for a given worker
        """
        queryset = self.queryset
        worker = self.request.query_params.get('worker')
        if worker is not None:
            # worker page filters and reverses the order
            queryset = queryset.filter(worker_id=worker)
        return queryset



from yawn.settings.base import *  # NOQA

SECRET_KEY = 'example secret key, change me'
DEBUG = True

# disable SSL settings
CSRF_COOKIE_SECURE = False
SECURE_PROXY_SSL_HEADER = None
SECURE_SSL_REDIRECT = False

# anonymous read-only allowed
REST_FRAMEWORK['DEFAULT_PERMISSION_CLASSES'] = [  # noqa
    'rest_framework.permissions.IsAuthenticatedOrReadOnly',
]



"""
This helper command will import any python callable on your python path and
call it with the supplied arguments.

Use `yawn.task.decorators.make_task` to
"""
import importlib

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Execute a python callable'

    def add_arguments(self, parser):
        parser.add_argument('module', help='The python module to import, i.e. animal.bird')
        parser.add_argument('callable', help='The python callable to invoke, i.e. Swallow')
        parser.add_argument('argument', nargs='*', help='Arguments to pass to the callable')

    def handle(self, *args, **options):
        self.stdout.write('Importing module %s' % options['module'])
        module_ = importlib.import_module(options['module'])

        arguments = ''
        if options['argument']:
            arguments = "'{}'".format("', '".join(options['argument']))

        self.stdout.write('Calling %s(%s)' % (options['callable'], arguments))
        getattr(module_, options['callable'])(*options['argument'])

        self.stdout.write('Execution complete')



from rest_framework.permissions import BasePermission


class ModelPermissions(BasePermission):
    """Allow authenticated users to read, and staff to write"""

    def has_permission(self, request, view):
        # Workaround to ensure DjangoModelPermissions are not applied
        # to the root view when using DefaultRouter.
        if getattr(view, '_ignore_model_permissions', False):
            return True

        if not request.user.is_authenticated:
            return False

        if request.user.is_staff:
            return True

        # authenticated, but not staff: only allow read
        return request.method in ('GET', 'HEAD', 'OPTIONS')



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-01-22 20:00
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.functions
import yawn.utilities.cron


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Execution',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.TextField(
                    choices=[('running', 'running'), ('succeeded', 'succeeded'), ('failed', 'failed'),
                             ('killed', 'killed'), ('lost', 'lost')], default='running')),
                ('start_timestamp', models.DateTimeField(default=django.db.models.functions.Now)),
                ('stop_timestamp', models.DateTimeField(null=True)),
                ('exit_code', models.IntegerField(null=True)),
                ('stdout', models.TextField(blank=True, default='')),
                ('stderr', models.TextField(blank=True, default='')),
            ],
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
            ],
        ),
        migrations.CreateModel(
            name='Queue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=True, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='Run',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('submitted_time', models.DateTimeField()),
                ('scheduled_time', models.DateTimeField(null=True)),
                ('status',
                 models.TextField(choices=[('running', 'running'), ('succeeded', 'succeeded'), ('failed', 'failed')],
                                  default='running')),
                ('parameters', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.TextField(
                    choices=[('waiting', 'waiting'), ('queued', 'queued'), ('running', 'running'),
                             ('succeeded', 'succeeded'), ('failed', 'failed'), ('upstream_failed', 'upstream_failed')],
                    default='waiting')),
                ('run', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Run')),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=True, db_index=False)),
                ('command', models.TextField()),
                ('max_retries', models.IntegerField(default=0)),
                ('timeout', models.IntegerField(null=True)),
                ('queue', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Queue')),
                ('upstream', models.ManyToManyField(blank=True, related_name='downstream', to='yawn.Template')),
            ],
        ),
        migrations.CreateModel(
            name='Worker',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
                ('status', models.TextField(choices=[('active', 'active'), ('exited', 'exited'), ('lost', 'lost')],
                                            default='active')),
                ('start_timestamp', models.DateTimeField(default=django.db.models.functions.Now)),
                ('last_heartbeat', models.DateTimeField(default=django.db.models.functions.Now)),
            ],
        ),
        migrations.CreateModel(
            name='Workflow',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', models.IntegerField(editable=False)),
                ('schedule_active', models.BooleanField(default=False)),
                ('schedule', models.TextField(null=True, validators=[yawn.utilities.cron.cron_validator])),
                ('next_run', models.DateTimeField(null=True)),
                ('parameters', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
            ],
        ),
        migrations.CreateModel(
            name='WorkflowName',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=True, unique=True)),
                ('current_version',
                 models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='is_current',
                                      to='yawn.Workflow')),
            ],
        ),
        migrations.AddField(
            model_name='workflow',
            name='name',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.WorkflowName'),
        ),
        migrations.AddField(
            model_name='template',
            name='workflow',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.PROTECT, to='yawn.Workflow'),
        ),
        migrations.AddField(
            model_name='task',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Template'),
        ),
        migrations.AddField(
            model_name='run',
            name='workflow',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Workflow'),
        ),
        migrations.AddField(
            model_name='message',
            name='queue',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Queue'),
        ),
        migrations.AddField(
            model_name='message',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Task'),
        ),
        migrations.AddField(
            model_name='execution',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Task'),
        ),
        migrations.AddField(
            model_name='execution',
            name='worker',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='yawn.Worker'),
        ),
        migrations.AlterUniqueTogether(
            name='workflow',
            unique_together=set([('name', 'version')]),
        ),
        migrations.AlterUniqueTogether(
            name='template',
            unique_together=set([('workflow', 'name')]),
        ),
    ]



from datetime import timedelta

from django.db import models
from django.db.models import functions

from yawn.utilities import logger


class Worker(models.Model):
    """Information about current and past workers"""
    #
    # NOTE: consider instead taking an advisory lock for each worker,
    # and using it to check if a worker is still connected.
    # See `pg_try_advisory_lock` and `select * from pg_locks where locktype = 'advisory'`
    # That would give more immediate feedback, but its not clear we need to be faster.
    #
    ACTIVE = 'active'
    EXITED = 'exited'
    LOST = 'lost'
    STATUS_CHOICES = [(x, x) for x in (ACTIVE, EXITED, LOST)]

    name = models.TextField(blank=False)
    status = models.TextField(choices=STATUS_CHOICES, default=ACTIVE)
    start_timestamp = models.DateTimeField(default=functions.Now)
    last_heartbeat = models.DateTimeField(default=functions.Now)

    @staticmethod
    def find_lost(timeout):
        from yawn.task.models import Execution

        # Make a sparse index so looking up active workers is fast:
        # CREATE INDEX yawn_worker_active ON yawn_worker (status) WHERE status = 'active'
        lost = Worker.objects.filter(
            status=Worker.ACTIVE, last_heartbeat__lt=functions.Now() - timedelta(seconds=timeout)
        )
        for worker in lost:
            logger.warning('Marking %r as lost', worker)
            worker.status = Worker.LOST
            worker.save()

            executions = worker.execution_set.filter(status=Execution.RUNNING)

            for execution in executions:
                logger.warning('Marking %r as lost', execution)
                execution.mark_finished(lost=True)

    def __str__(self):
        return self.name


class Queue(models.Model):
    """Arbitrary tag defining where tasks run."""

    name = models.TextField(unique=True)

    _default = None

    def __str__(self):
        return self.name

    @classmethod
    def get_default_queue(cls):
        if not cls._default:
            cls._default = Queue.objects.get_or_create(name='default')[0]
        return cls._default


class Message(models.Model):
    """The order of tasks waiting to be processed, like messages on a queue"""

    # I hope we never get to 9 Quintillion (9,223,372,036,854,775,807) messages
    id = models.BigAutoField(primary_key=True)

    queue = models.ForeignKey(Queue, models.PROTECT)
    task = models.ForeignKey('Task', models.PROTECT)



from whitenoise.middleware import WhiteNoiseMiddleware


class DefaultFileMiddleware(WhiteNoiseMiddleware):
    """
    Serve index.html if not /api/ or a static file

    This allows users to navigate directly to a URL like http://127.0.0.1:8000/workflows/2
    and we return index.html, and then react-router interprets the path.
    """

    def process_request(self, request):

        if request.path_info.startswith('/api'):
            return  # handled by Django

        static_file = self.files.get(request.path_info)

        if static_file is None:
            # serve the homepage on non-existent file
            static_file = self.files.get('/index.html')

        return self.serve(static_file, request)



import shlex

from yawn.task.models import Template, Task
from yawn.workflow.models import Workflow  # noqa - needed to load the foreign key!
from yawn.worker.models import Queue


def delay(func, *args, timeout=None, max_retries=0, queue=None):
    arguments = [shlex.quote(arg) for arg in args]
    command = 'yawn exec {0.__module__} {0.__name__} {1}'.format(
        func, ' '.join(arguments)).strip()
    task_name = '{0.__module__}.{0.__name__}({1})'.format(
        func, ', '.join(arguments))

    if queue:
        queue_obj, _ = Queue.objects.get_or_create(name=queue)
    else:
        queue_obj = Queue.get_default_queue()

    template, _ = Template.objects.get_or_create(
        name=task_name,
        command=command,
        queue=queue_obj,
        max_retries=max_retries,
        timeout=timeout
    )
    task = Task.objects.create(
        template=template
    )
    task.enqueue()
    return task



import random

from django.core.management.base import BaseCommand

from yawn.task.models import Template
from yawn.workflow.models import WorkflowName
from yawn.workflow.serializers import WorkflowSerializer
from yawn.workflow.tests.utils import load_sample_workflow


class Command(BaseCommand):
    help = 'Load sample data'

    def handle(self, *args, **options):
        self.stdout.write('Creating sample workflows!')

        self.stdout.write('First, a simple four task workflow...')
        name, _ = WorkflowName.objects.get_or_create(name='Simple Workflow Example')
        workflow = name.new_version(parameters={'MY_OBJECT_ID': '1', 'SOME_SETTING': 'false'})
        task1 = Template.objects.create(workflow=workflow, name='start', command='echo Starting...')
        task2 = Template.objects.create(workflow=workflow, name='task2', command='echo Working on $MY_OBJECT_ID')
        task2.upstream.add(task1)
        task3 = Template.objects.create(workflow=workflow, name='task3',
                                        command='yawn exec yawn.management.tests.test_exec '
                                                'ExampleClass "thing to log"')
        task3.upstream.add(task1)
        task4 = Template.objects.create(workflow=workflow, name='done', command='echo Finished!')
        task4.upstream.add(task2, task3)

        workflow.submit_run(parameters={'child': 'true'})
        workflow.submit_run(parameters={'child': 'true'})
        workflow.submit_run(parameters={'child': 'true'})

        self.stdout.write('Second, a big beast workflow...')

        name, _ = WorkflowName.objects.get_or_create(name='Big Ol\' Workflow Example')
        workflow = name.new_version()
        tasks = []
        for x in range(30):
            task = Template.objects.create(workflow=workflow, name='task%s' % x, command='echo stuff')
            if tasks:
                task.upstream.add(*random.sample(tasks, k=min(x, 5)))
            tasks.append(task)

        # Create from a dictionary:
        self.stdout.write('Third, an example using the serializer...')
        serializer = WorkflowSerializer(data=load_sample_workflow())
        serializer.is_valid(raise_exception=True)
        workflow = serializer.save()
        workflow.submit_run()



from django.db.models import Count
from rest_framework import viewsets

from yawn.worker.models import Worker, Queue
from yawn.worker.serializers import WorkerSerializer, QueueSerializer


class WorkerViewSet(viewsets.GenericViewSet,
                    viewsets.mixins.ListModelMixin,
                    viewsets.mixins.RetrieveModelMixin):
    """
    Worker endpoint, GET(list)
    """
    queryset = Worker.objects.all().order_by('-id')

    serializer_class = WorkerSerializer


class QueueViewSet(viewsets.GenericViewSet,
                   viewsets.mixins.ListModelMixin,
                   viewsets.mixins.RetrieveModelMixin,
                   viewsets.mixins.UpdateModelMixin):
    """
    Queue endpoint GET(list,detail), PATCH to clear all messages
    """
    queryset = Queue.objects.all().annotate(Count('message')).order_by('id')

    serializer_class = QueueSerializer



import pytest
from django.contrib.auth.models import User


@pytest.fixture()
def user():
    return User.objects.get(username='test_user')


def test_list_users(client, user):
    response = client.get('/api/users/')
    assert response.status_code == 200
    users = response.data['results']
    assert len(users) == 1
    assert users[0]['email'] == user.email
    assert users[0]['username'] == user.username


def test_get_me(client, user):
    response = client.get('/api/users/me/')
    assert response.status_code == 200
    assert response.data['email'] == user.email
    assert response.data['username'] == user.username


def test_update_token(client, user):
    response = client.patch('/api/users/%s/' % user.id, {'refresh_token': True, 'username': 'other'})
    assert response.status_code == 200, response.data
    assert response.data['api_token'] is not None
    assert response.data['api_token'] == user.auth_token.key
    assert response.data['username'] == 'other'


def test_login_success(client, user):
    user.set_password('abcd')
    user.save()
    response = client.patch('/api/users/login/', {'username': user.username, 'password': 'abcd'})
    assert response.status_code == 200, response.data


def test_login_failure(client, user):
    response = client.patch('/api/users/login/', {'username': user.username, 'password': 'abcd'})
    assert response.status_code == 401, response.data



"""
Test workflow logic
A
 \
  C - D
 /
B

Tasks:
A succeeded, do nothing to C(waiting)
A & B succeed, submit C(waiting)
A & B succeed, submit C(upstream_failed)
A & B succeed, don't submit C(failed)
A succeeded but B failed, mark C and D upstream_failed
"""
from yawn.worker.models import Queue, Worker
from yawn.task.models import Task, Execution, Template


def test_first_queued(run):
    task1, task2 = run.task_set.all()
    task2.enqueue()
    queue = Queue.objects.get(name='default')
    first = Task.first_queued([queue.id])
    assert first.id == task1.id
    first = Task.first_queued([queue.id])
    assert first.id == task2.id
    first = Task.first_queued([queue.id])
    assert first is None


def test_first_queued_by_queue(run):
    task1, task2 = run.task_set.all()
    default_queue = Queue.objects.get(name='default')
    other_queue = Queue.objects.create(name='test')
    task2.template.queue = other_queue
    task2.enqueue()
    first = Task.first_queued([default_queue.id])
    assert first.id == task1.id
    first = Task.first_queued([default_queue.id])
    assert first is None
    first = Task.first_queued([other_queue.id])
    assert first.id == task2.id


def test_start_execution(run):
    task1, task2 = run.task_set.all()
    worker = Worker.objects.create(name='worker1')
    execution = task1.start_execution(worker)
    assert execution.status == Execution.RUNNING
    assert task1.status == Task.RUNNING


def test_execution_exitcode(run):
    task1, task2 = run.task_set.all()
    worker = Worker.objects.create(name='worker1')
    execution = task1.start_execution(worker)

    execution.mark_finished(exit_code=0)
    task2.refresh_from_db()
    assert execution.status == Execution.SUCCEEDED
    assert task1.status == Task.SUCCEEDED
    assert task2.status == Task.QUEUED

    task2.status = Task.WAITING
    task2.save()

    execution.mark_finished(exit_code=1)
    task2.refresh_from_db()
    assert execution.status == Execution.FAILED
    assert task1.status == Task.FAILED
    assert task2.status == Task.UPSTREAM_FAILED

    execution.mark_finished(lost=True)
    assert execution.status == Execution.LOST
    assert task1.status == Task.QUEUED


def test_execution_output(run):
    task1, task2 = run.task_set.all()
    worker = Worker.objects.create(name='worker1')
    execution = task1.start_execution(worker)

    execution.update_output(execution.id, 'foo', 'blah')
    execution.refresh_from_db()
    assert execution.stdout == 'foo'
    assert execution.stderr == 'blah'

    # only provide one, check it also works
    execution.update_output(execution.id, 'bar', None)
    execution.refresh_from_db()
    assert execution.stdout == 'foobar'
    assert execution.stderr == 'blah'


def test_task_without_workflow():
    template = Template.objects.create(name='task1', command='')
    task = Task.objects.create(template=template)
    worker = Worker.objects.create(name='worker1')

    # asserts that a task can be run without having a workflow associated
    execution = task.start_execution(worker)
    execution.mark_finished(exit_code=0)
    task.refresh_from_db()
    assert task.status == Task.SUCCEEDED



from yawn.task.models import Template, Task, Execution  # NOQA
from yawn.worker.models import Worker
from yawn.workflow.models import WorkflowName


def test_find_lost():
    name = WorkflowName.objects.create(name='workflow1')
    workflow = name.new_version(parameters={'parent': True, 'child': False})
    task1 = Template.objects.create(workflow=workflow, name='task1', command=[''])
    task2 = Template.objects.create(workflow=workflow, name='task2', command=[''])
    task2.upstream.add(task1)
    run = workflow.submit_run(parameters={'child': True})
    task1, task2 = run.task_set.all()  # type: Task
    worker = Worker.objects.create(name='worker1')
    execution = task1.start_execution(worker)

    worker.find_lost(-1)
    worker.refresh_from_db()
    execution.refresh_from_db()

    assert worker.status == Worker.LOST
    assert execution.status == Execution.LOST



import pytest

from django.utils.dateparse import parse_datetime

from yawn.workflow.models import WorkflowName
from yawn.workflow.tests.utils import load_sample_workflow

data = pytest.fixture(load_sample_workflow)


def test_create_workflow(client, data):
    response = client.post('/api/workflows/', data)
    assert response.status_code == 201
    assert response.data['name'] == data['name']
    assert response.data['next_run'] is not None
    assert response.data['schedule'] == data['schedule']
    assert response.data['schedule_active'] == data['schedule_active']
    assert response.data['parameters'] == data['parameters']
    assert response.data['tasks'][0]['command'] == data['tasks'][0]['command']
    assert response.data['tasks'][0]['queue'] == 'default'
    assert response.data['tasks'][0]['timeout'] == data['tasks'][0]['timeout']
    assert response.data['tasks'][1]['upstream'] == data['tasks'][1]['upstream']
    assert response.data['tasks'][1]['max_retries'] == 0  # default
    assert response.data['tasks'][1]['timeout'] is None  # default


def test_create_workflow_versions(client, data):
    response = client.post('/api/workflows/', data)
    assert response.status_code == 201, response.data
    assert response.data['version'] == 1

    # no change, same version returned
    response = client.post('/api/workflows/', data)
    assert response.data['version'] == 1

    # any change, new version created
    data['tasks'][0]['max_retries'] = 2
    response = client.post('/api/workflows/', data)
    assert response.status_code == 201, response.data
    assert response.data['version'] == 2


def test_invalid_fields(client, data):
    data['name'] = ''
    data['parameters'] = 'not a dict'
    data['tasks'][1]['upstream'].append('invalid_task')
    response = client.post('/api/workflows/', data)
    assert response.status_code == 400, response.data
    assert 'This field may not be blank' in response.data['name'][0]
    assert 'must be a dictionary' in response.data['parameters'][0]
    assert 'upstream task(s) invalid_task' in response.data['tasks'][0]


def test_more_invalid_fields(client, data):
    data['parameters'] = {'invalid variable': 1}
    data['tasks'] = []
    response = client.post('/api/workflows/', data)
    assert response.status_code == 400, response.data
    assert 'Invalid parameter key' in response.data['parameters'][0]
    assert 'Invalid parameter value' in response.data['parameters'][1]
    assert 'non_field_errors' in response.data['tasks']


def test_list_workflow_names(client):
    name = WorkflowName.objects.create(name='workflow1')
    version1 = name.new_version()
    version2 = name.new_version()
    response = client.get('/api/names/')
    assert response.status_code == 200, response.data
    assert response.data['results'][0]['name'] == name.name
    response = client.get('/api/names/{}/'.format(name.id))
    assert response.status_code == 200, response.data
    assert response.data['current_version_id'] == version2.id
    assert response.data['current_version'] == version2.version
    assert response.data['task_count'] == 0
    assert response.data['versions'] == [version1.id, version2.id]


def test_get_run(client, run):
    url = '/api/runs/%s/' % run.id
    response = client.get(url)
    run.refresh_from_db()  # to get submitted_time which is from the DB
    assert response.status_code == 200, response.data
    assert response.data['status'] == run.status
    assert response.data['parameters'] == run.parameters
    assert response.data['scheduled_time'] == run.scheduled_time
    assert parse_datetime(response.data['submitted_time']) == run.submitted_time
    tasks = list(run.task_set.all())
    # we get all the tasks in the list view as they are used in the DetailHistory table
    assert response.data['tasks'][0]['status'] == tasks[0].status
    assert response.data['tasks'][1]['status'] == tasks[1].status


def test_post_run(client, run):
    """create a new run of the workflow"""
    url = '/api/runs/'
    workflow = run.workflow
    data = {'workflow_id': workflow.id, 'parameters': {'A': '1', 'B': 'false'}}
    response = client.post(url, data)
    assert response.status_code == 201, response.data
    assert response.data['id'] == run.id + 1
    assert response.data['status'] == run.RUNNING
    merged_parameters = workflow.parameters.copy()
    merged_parameters.update(data['parameters'])
    assert response.data['parameters'] == merged_parameters

    # without parameters
    data = {'workflow_id': workflow.id}
    response = client.post(url, data)
    assert response.status_code == 201, response.data
    assert response.data['id'] == run.id + 2
    assert response.data['status'] == run.RUNNING
    assert response.data['parameters'] == workflow.parameters


def test_patch_run(client, run):
    url = '/api/runs/%s/' % run.id
    update = {'parameters': {'A': '1', 'B': 'false'}}
    response = client.patch(url, update)
    run.refresh_from_db()
    assert response.status_code == 200, response.data
    assert response.data['parameters'] == update['parameters']
    assert response.data['parameters'] == run.parameters



from django.contrib.auth.models import User
from django.contrib.auth import authenticate, login, logout
from rest_framework import viewsets, status
try:
    from rest_framework.decorators import action
except ImportError:
    from rest_framework.decorators import list_route as action
from rest_framework.response import Response
from rest_framework.permissions import AllowAny
from rest_framework.settings import api_settings

from yawn.user.permissions import ModelPermissions
from .serializers import UserSerializer, LoginSerializer


class UserViewSet(viewsets.GenericViewSet,
                  viewsets.mixins.ListModelMixin,
                  viewsets.mixins.CreateModelMixin,
                  viewsets.mixins.RetrieveModelMixin,
                  viewsets.mixins.UpdateModelMixin):
    """
    User endpoint, GET(list, detail), PATCH to change

    This ViewSet has `permission_classes` set and `/me/` has the default permissions
    so that the default permission class can be set to IsAuthenticatedOrReadOnly
    without leaking user API tokens.
    """
    queryset = User.objects.all().order_by('id')

    serializer_class = UserSerializer
    permission_classes = (ModelPermissions,)

    @action(methods=['get'], permission_classes=api_settings.DEFAULT_PERMISSION_CLASSES, detail=False)
    def me(self, request):
        serializer = self.get_serializer(request.user)
        return Response(serializer.data)

    @action(methods=['patch'], permission_classes=[AllowAny], detail=False)
    def login(self, request):
        credentials = LoginSerializer(data=request.data)
        credentials.is_valid(raise_exception=True)
        user = authenticate(request, **credentials.data)
        if not user:
            return Response({'detail': 'Login failed'}, status.HTTP_401_UNAUTHORIZED)

        login(request, user)
        return Response({'detail': 'Login succeeded'})

    @action(methods=['delete'], detail=False)
    def logout(self, request):
        logout(request)
        return Response({'detail': 'Logout succeeded'})



from collections import OrderedDict

from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response


class AdjustablePagination(PageNumberPagination):
    page_size_query_param = 'page_size'
    max_page_size = 1000

    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('count', self.page.paginator.count),
            ('page', self.page.number),
            ('page_size', self.get_page_size(self.request)),
            ('page_count', self.page.paginator.num_pages),
            ('results', data)
        ]))



from rest_framework import serializers

from yawn.task.models import Task, Execution
from yawn.worker.serializers import MessageSerializer, WorkerSerializer
from yawn.workflow.models import Workflow


class SimpleWorkflowSerializer(serializers.ModelSerializer):
    name = serializers.CharField(source='name.name', read_only=True)

    class Meta:
        model = Workflow
        fields = ('id', 'name', 'version')


class TaskSerializer(serializers.ModelSerializer):
    name = serializers.CharField(source='template.name', read_only=True)
    workflow = SimpleWorkflowSerializer(source='template.workflow', read_only=True)

    class Meta:
        model = Task
        exclude = ('run', 'template')


class ExecutionDetailSerializer(serializers.ModelSerializer):
    worker = WorkerSerializer(read_only=True)

    class Meta:
        model = Execution
        exclude = ('task',)


class ExecutionListSerializer(serializers.ModelSerializer):
    worker = WorkerSerializer(read_only=True)
    task = TaskSerializer(read_only=True)
    minutes_running = serializers.SerializerMethodField()

    class Meta:
        model = Execution
        fields = ('id', 'task', 'worker', 'status', 'start_timestamp', 'minutes_running')

    def get_minutes_running(self, obj):
        if obj.stop_timestamp:
            runtime = (obj.stop_timestamp - obj.start_timestamp).total_seconds()
            return '{:.0f}m {:.2f}s'.format(runtime // 60, runtime % 60)


class TaskDetailSerializer(TaskSerializer):
    executions = serializers.SerializerMethodField()
    messages = MessageSerializer(many=True, source='message_set', read_only=True)

    max_retries = serializers.IntegerField(source='template.max_retries')
    timeout = serializers.IntegerField(source='template.timeout')
    command = serializers.CharField(source='template.command')

    # actions
    terminate = serializers.IntegerField(write_only=True)
    enqueue = serializers.BooleanField(write_only=True)

    def get_executions(self, instance):
        executions = instance.execution_set.order_by('id')
        return ExecutionDetailSerializer(executions, many=True).data

    def update(self, instance, validated_data):
        if validated_data.get('terminate'):
            instance.execution_set.filter(
                id=validated_data['terminate'], status=Execution.RUNNING
            ).update(status=Execution.KILLED)
        if validated_data.get('enqueue'):
            instance.enqueue()
        return instance



import shlex

from gunicorn.app.base import BaseApplication
from django.core.management.base import BaseCommand
from django.core.wsgi import get_wsgi_application


class WSGIApplication(BaseApplication):
    """A Gunicorn Application, with logic to parse config passed from the command"""

    def __init__(self, options):
        self.options = options
        super().__init__()

    def load_config(self):
        parser = self.cfg.parser()

        # by default, log access to stdout
        self.cfg.set('accesslog', '-')

        env_vars = self.cfg.get_cmd_args_from_env()
        if env_vars:
            env_args = parser.parse_args(env_vars)
            self._set_cfg(env_args)

        if self.options:
            cmd_args = parser.parse_args(shlex.split(self.options))
            self._set_cfg(cmd_args)

    def _set_cfg(self, args):
        """Given an argparse namespace, set cfg variables"""
        for k, v in vars(args).items():
            if v is None:
                continue
            if k == "args":
                continue
            self.cfg.set(k.lower(), v)

    def load(self):
        return get_wsgi_application()


class Command(BaseCommand):
    help = 'Run the webserver.'
    requires_migrations_checks = True

    def add_arguments(self, parser):
        """Adds the Gunicorn command line args to YAWN"""
        parser.add_argument('--gunicorn', type=str,
                            help='Arguments to pass to gunicorn. See `gunicorn -h` for documentation. '
                                 'Alternatively put them in the GUNICORN_CMD_ARGS env variable.')

    def handle(self, *args, **options):
        print('  Invoking gunicorn YAWN webserver')

        # Run Gunicorn
        gunicorn_args = options.get('gunicorn')
        app = WSGIApplication(gunicorn_args)
        app.run()



import os

# Production settings that probably won't work with local dev,
# but typically desirable when deployed.

# You also need to set SECRET_KEY and ALLOWED_HOSTS!!

# Tell the browser only to send the cookie with HTTPS requests
CSRF_COOKIE_SECURE = True

# Tell Django that it is running behind a proxy and should forward HTTP->HTTPS
SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')
SECURE_SSL_REDIRECT = True
SECURE_REDIRECT_EXEMPT = ['^api/healthy/$']

# Custom YAWN settings

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': 'yawn',
        'HOST': 'localhost'  # not using unix socket b/c circle build
    }
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'rest_framework.authtoken',
    'yawn',
]

REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': (
        'yawn.user.permissions.ModelPermissions',
    ),
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_PAGINATION_CLASS': 'yawn.utilities.pagination.AdjustablePagination',
    'PAGE_SIZE': 100,
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
}

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Application definition

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'yawn.utilities.middleware.DefaultFileMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'yawn.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.10/howto/static-files/

STATIC_URL = '/static/'

STATIC_ROOT = os.path.join(BASE_DIR, 'staticfiles', 'static')

WHITENOISE_ROOT = os.path.join(BASE_DIR, 'staticfiles')

# interferes with the DefaultFileServer
WHITENOISE_AUTOREFRESH = False
WHITENOISE_USE_FINDERS = False



from .debug import *  # NOQA

ALLOWED_HOSTS = ['testserver']

# test raven integration:

# INSTALLED_APPS += ['raven.contrib.django']
#
# RAVEN_CONFIG = {
#     'dsn': os.environ.get('SENTRY_DSN'),
# }



from unittest import mock

from django.core import management

from yawn.management.commands import worker


@mock.patch.object(worker.Main, 'run')
def test_worker(mock_run):
    management.call_command('worker')
    assert mock_run.called



import time
import pytest

from unittest import mock

from .. import executor


@pytest.fixture()
def manager():
    return executor.Manager()


def test_read_output(manager):
    execution_id = 1
    command = 'echo starting && sleep 0.1 && echo stopping 1>&2'
    manager.start_subprocess(execution_id, command, environment={}, timeout=None)

    # stdout is ready
    results = manager.read_output(timeout=1)
    assert len(results) == 1, 'Unexpected state {}'.format(manager.running[1])
    assert results[0].stdout == 'starting\n'
    assert results[0].stderr is None
    assert results[0].returncode is None
    assert results[0].execution_id == execution_id

    # sleep to ensure the returncode is ready
    time.sleep(1)

    # stderr is ready
    results = manager.read_output(timeout=1)
    assert len(results) == 1, 'Unexpected state {}'.format(manager.running[1])
    assert results[0].stdout is None
    assert results[0].stderr == 'stopping\n'
    assert results[0].execution_id == execution_id

    # grhh... sometime it exits later, on circleci it often exits concurrently
    if results[0].returncode is None:
        results = manager.read_output(timeout=1)
        assert len(results) == 1, 'Unexpected state {}'.format(manager.running[1])

    assert results[0].execution_id == execution_id
    assert results[0].returncode == 0


def test_timeout(manager):
    command = 'sleep 10'
    manager.start_subprocess(1, command, environment={}, timeout=-1)

    results = manager.read_output()  # first read kills, but the pipes are still open
    assert results == []

    results = manager.read_output(timeout=1)
    assert len(results) == 1, 'Unexpected state {}'.format(manager.running[1])
    assert results[0].stdout is None
    assert results[0].stderr is None
    assert results[0].returncode == -9


@mock.patch('yawn.worker.executor.os.killpg')
def test_timeout_already_exited(mock_kill, manager):
    command = 'true'
    manager.start_subprocess(1, command, environment={}, timeout=-1)

    results = manager.read_output(timeout=1)
    assert len(results) == 1, 'Unexpected state {}'.format(manager.running[1])
    assert results[0].stdout is None
    assert results[0].stderr is None
    assert results[0].returncode == 0

    assert mock_kill.call_count == 1


def test_mark_terminated(manager):
    execution_id = 1
    execution = executor.Execution(None, execution_id, None)
    manager.running[execution_id] = execution
    manager.mark_terminated([1, 2])
    assert execution.deadline <= time.monotonic()



import datetime

from django.core import validators


class Crontab:
    """
    Simplified Crontab

    Support "minute hour weekday" components of a standard cron job.
    - "*/15 2,7,15 1-5" means "every fifteen minutes, on hours 2 7 15, Monday-Friday"
    - Minutes are from 0-59, hours from 0-23, and days from 0(Sunday)-6(Saturday)
    - Fields can contain multiple comma-separated values
    - Values can be an integer or repeating pattern of the '*/2' variety
    """

    def __init__(self, schedule: str):
        self.schedule = schedule
        components = schedule.split(' ')
        if len(components) != 3:
            raise ValueError('Crontab must be three space-delimited components')

        minutes, hours, weekdays = components
        self.minutes = parse(minutes, 60)
        self.hours = parse(hours, 24)
        self.weekdays = parse(weekdays, 24)

    def __repr__(self):
        return '<Crontab: {}>'.format(self.schedule)

    def next_run(self, current_time: datetime.datetime) -> datetime.datetime:
        """Given the current time, when is the next scheduled run?"""
        # if next run is next day, get smallest hour, smallest minute
        # if next run is today, future hour, get smallest minute
        # if next run is today, this hour, get next greatest minute
        next_run = datetime.datetime(current_time.year, current_time.month, current_time.day,
                                     tzinfo=current_time.tzinfo)
        weekday = current_time.isoweekday()
        weekday = 0 if weekday == 7 else weekday  # Move Sunday to day 0
        if weekday in self.weekdays:
            # could be a run today
            if current_time.hour in self.hours:
                # could be a run this hour
                for minute in self.minutes:
                    if minute > current_time.minute:
                        # there is a run this hour
                        return next_run.replace(hour=current_time.hour, minute=minute)
            # no run this hour, check future hours
            for hour in self.hours:
                if hour > current_time.hour:
                    # there is a run today
                    return next_run.replace(hour=hour, minute=self.minutes[0])
        # no run today, look for next matching weekday
        for day in range(1, 7):
            next_run += datetime.timedelta(days=1)
            weekday = next_run.isoweekday()
            weekday = 0 if weekday == 7 else weekday  # Move Sunday to day 0
            if weekday in self.weekdays:
                return next_run.replace(hour=self.hours[0], minute=self.minutes[0])
        raise RuntimeError('No next run found for schedule {}'.format(self.schedule))


def parse(pattern: str, max_value: int):
    """Convert a string crontab component into a set of integers less than a given max"""
    values = set()
    for part in pattern.split(','):

        fraction = part.split('/')
        if len(fraction) == 1:
            numerator, denominator = part, None
        elif len(fraction) == 2:
            numerator, denominator = fraction[0], int(fraction[1])
        else:
            raise ValueError('Expression {} should contain zero or one slash (/)'.format(part))

        if numerator == '*':
            lower, upper = 0, max_value
        elif '-' in numerator:
            lower, upper = numerator.split('-')
            lower, upper = int(lower), int(upper) + 1
        else:
            lower, upper = int(numerator), int(numerator) + 1

        if lower < 0 or upper > max_value:
            raise ValueError('Expression {} is outside the range {}-{}'.format(
                part, 0, max_value))

        for x in range(lower, upper):
            if denominator is None or x % denominator == 0:
                values.add(x)

    if not values:
        raise ValueError('Expression {} gives no runs'.format(pattern))

    return sorted(values)


def cron_validator(crontab: str):
    try:
        Crontab(crontab)
    except ValueError as exc:
        raise validators.ValidationError(
            'Invalid crontab expression: {} ({})'.format(crontab, exc))



from django.conf.urls import include, url
from django.http import HttpResponse
from rest_framework import routers

from yawn.task.views import TaskViewSet, ExecutionViewSet
from yawn.worker.views import QueueViewSet, WorkerViewSet
from yawn.workflow.views import WorkflowViewSet, WorkflowNameViewSet, RunViewSet
from yawn.user.views import UserViewSet

router = routers.DefaultRouter()
router.include_format_suffixes = False
router.register(r'names', WorkflowNameViewSet)
router.register(r'workflows', WorkflowViewSet)
router.register(r'runs', RunViewSet)
router.register(r'tasks', TaskViewSet)
router.register(r'queues', QueueViewSet)
router.register(r'workers', WorkerViewSet)
router.register(r'executions', ExecutionViewSet)
router.register(r'users', UserViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browsable API.
urlpatterns = [
    url(r'^api/', include(router.urls)),
    url(r'^api/healthy/$', lambda x: HttpResponse('ok')),
]



import functools

from django import db
from django.db import OperationalError

from yawn.utilities import logger


def close_on_exception(func):
    """
    A wrapper to close the database connection if a DB error occurs,
    so that it will get re-opened on the next use.

    Squashes the exception and logs it.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except OperationalError:
            logger.error('Database error, closing connection', exc_info=True)
            db.connection.close()
            assert db.connection.closed_in_transaction is False, \
                'Could not close connection, probably because this wrapper ' \
                'was used inside an transaction.atomic() block.'

    return wrapper


def current_time():
    """Return the database time"""
    with db.connection.cursor() as cursor:
        cursor.execute("SELECT STATEMENT_TIMESTAMP()")
        row = cursor.fetchone()
    return row[0]



from yawn.task.helpers import delay


def some_function(*args):
    pass


class SomeClass:
    pass


def test_quoted_arg():
    task = delay(SomeClass, 'A small "taste" of chaos')
    assert task.template.command == 'yawn exec yawn.task.tests.test_helpers ' \
                                    'SomeClass \'A small "taste" of chaos\''


def test_many_args():
    task = delay(some_function, 'something', '1')
    assert task.template.command == "yawn exec yawn.task.tests.test_helpers " \
                                    "some_function something 1"


def test_deduplication():
    task1 = delay(some_function)
    task2 = delay(some_function)
    assert task1.template_id == task2.template_id


def test_queued():
    task = delay(SomeClass, queue='queue', max_retries=2, timeout=10)
    assert task.message_set.count() == 1
    assert task.message_set.first().queue.name == 'queue'
    assert task.template.max_retries == 2
    assert task.template.timeout == 10



from django.db.models import Count
from rest_framework import viewsets

from yawn.workflow.serializers import WorkflowSerializer, WorkflowNameSerializer, RunSerializer
from yawn.workflow.models import Workflow, WorkflowName, Run


class WorkflowNameViewSet(viewsets.ReadOnlyModelViewSet):
    """
    WorkflowName endpoint. Provides the name, current version number,
    current_version_id, and the select_version_id if a version was
    requested in the query string.

    Used to list the workflows and power version switcher.
    """
    queryset = WorkflowName.objects.select_related('current_version').annotate(
        task_count=Count('current_version__template')).order_by('id')
    page_size = 100000  # not much pagination

    serializer_class = WorkflowNameSerializer


class WorkflowViewSet(viewsets.GenericViewSet,
                      viewsets.mixins.ListModelMixin,
                      viewsets.mixins.RetrieveModelMixin,
                      viewsets.mixins.UpdateModelMixin,
                      viewsets.mixins.CreateModelMixin):
    """
    Workflow endpoint.

    The POST action is unusual as it checks for an existing, identical
    workflow and return it if found instead of creating a new object.
    """
    queryset = Workflow.objects.select_related('name').order_by('id')

    serializer_class = WorkflowSerializer


class RunViewSet(viewsets.GenericViewSet,
                 viewsets.mixins.ListModelMixin,
                 viewsets.mixins.RetrieveModelMixin,
                 viewsets.mixins.CreateModelMixin,
                 viewsets.mixins.UpdateModelMixin):
    """
    Run endpoint: GET(list,detail), POST(list) and PATCH(detail)
    """

    serializer_class = RunSerializer

    queryset = Run.objects.prefetch_related('task_set__template').order_by('id')

    def get_queryset(self):
        """
        Optionally filter to the runs for a given workflow
        """
        queryset = self.queryset
        workflow = self.request.query_params.get('workflow', None)
        if workflow is not None:
            queryset = queryset.filter(workflow_id=workflow)
        return queryset



import datetime
import signal
from unittest import mock

from django.utils import timezone

from yawn.task.models import Template, Task  # noqa: used in typing
from yawn.worker.executor import Result
from yawn.worker.models import Queue, Worker
from yawn.worker.main import Main, State
from yawn.workflow.models import WorkflowName, Workflow


def test_init_default_queue():
    worker = Main(1, 'test name', ['default'])
    assert worker.name == 'test name'
    assert worker.concurrency == 1
    assert len(worker.queue_ids) == 1
    queue = Queue.objects.get(id=worker.queue_ids[0])
    assert queue.name == 'default'


def test_init_custom_queues():
    queues = ['queue1', 'queue2']
    worker1 = Main(1, 'name', queues)
    worker2 = Main(1, 'name', queues)
    assert len(worker1.queue_ids) == 2
    # check that they end up with the same queue IDs
    assert sorted(worker1.queue_ids) == sorted(worker2.queue_ids)
    queue_names = Queue.objects.filter(
        id__in=worker1.queue_ids).values_list('name', flat=True)
    assert sorted(queue_names) == queues


@mock.patch('yawn.worker.main.time.sleep')
def test_run(mock_time):
    # so the worker exits immediately
    worker = Main(1, 'test name', ['default'])

    def set_shutdown(_):
        # stop the worker after one run
        worker.state = State.shutdown

    mock_time.side_effect = set_shutdown
    worker.run()


def test_schedule_workflows():
    name = WorkflowName.objects.create(name='workflow1')
    next_run = timezone.now() - datetime.timedelta(hours=1)
    workflow = Workflow.objects.create(
        name=name, version=1, schedule_active=True, schedule='0 0 *', next_run=next_run)
    worker = Main(1, 'test name', ['default'])
    worker.schedule_workflows()
    workflow.refresh_from_db()
    assert workflow.next_run > next_run


def test_signals():
    try:
        worker = Main(1, 'test name', ['default'])
        worker.handle_signals()
        worker.executor = mock.Mock()

        handler = signal.getsignal(signal.SIGINT)
        handler()
        assert worker.state == State.shutdown

        handler = signal.getsignal(signal.SIGTERM)
        handler()
        assert worker.state == State.terminate
        assert worker.executor.mark_terminated.called

    finally:
        # restore the default handlers
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)


def test_set_lost():
    worker = Main(1, 'test name', ['default'])
    worker.update_worker()  # first time creates the row
    worker.worker.status = Worker.LOST
    worker.worker.save()

    worker.executor = mock.Mock()
    worker.update_worker()

    assert worker.state == State.terminate
    assert worker.executor.mark_terminated.called


def test_save_results():
    # lots of setup
    worker = Main(1, 'test name', ['default'])
    worker.update_worker()

    name = WorkflowName.objects.create(name='workflow1')
    workflow = name.new_version(parameters={'parent': True, 'child': False})
    Template.objects.create(workflow=workflow, name='task1', command=[''])
    run = workflow.submit_run(parameters={'child': True})
    task1 = run.task_set.first()  # type: Task
    execution = task1.start_execution(worker.worker)

    result = Result(execution_id=execution.id)
    result.stdout = 'test stdout'
    worker.results.append(result)

    # trigger a save
    worker.save_results()

    execution.refresh_from_db()
    assert execution.stdout == result.stdout
    assert not worker.results



import io
import logging

import pytest
from django.core import management

LOG1 = 'In example_function'
LOG2 = 'In ExampleClass'


def example_function():
    logging.info(LOG1)


class ExampleClass:
    def __init__(self, argument):
        logging.basicConfig()
        logging.warning(argument)


def test_call_function(caplog):
    caplog.set_level(logging.INFO)
    stdout = io.StringIO()
    management.call_command('exec', 'yawn.management.tests.test_exec', 'example_function', stdout=stdout)
    assert 'Execution complete' in stdout.getvalue()
    assert caplog.records[0].message == LOG1


def test_call_class(caplog):
    stdout = io.StringIO()
    management.call_command('exec', 'yawn.management.tests.test_exec', 'ExampleClass', LOG2, stdout=stdout)
    assert 'Execution complete' in stdout.getvalue()
    assert caplog.records[0].message == LOG2


def test_invalid_module():
    stdout = io.StringIO()
    with pytest.raises(ImportError):
        management.call_command('exec', 'yawn.no_such_module', 'ExampleClass', stdout=stdout)



from django.contrib.auth.models import User
from rest_framework.authtoken.models import Token
from rest_framework import serializers


class UserSerializer(serializers.ModelSerializer):
    api_token = serializers.CharField(source='auth_token.key', read_only=True)
    refresh_token = serializers.BooleanField(write_only=True, default=False)
    password = serializers.CharField(write_only=True)

    class Meta:
        model = User
        fields = ('username', 'first_name', 'last_name', 'api_token', 'email',
                  'refresh_token', 'id', 'password', 'is_staff')

    def create(self, validated_data):
        validated_data.pop('refresh_token', None)
        return User.objects.create_user(**validated_data)

    def update(self, instance, validated_data):
        # support refreshing the token:
        if validated_data.pop('refresh_token', False):
            if hasattr(instance, 'auth_token'):
                instance.auth_token.delete()
            instance.auth_token = Token.objects.create(user=instance)
        # and setting the password, which causes an unhandled logout...
        if validated_data.get('password'):
            instance.set_password(validated_data.pop('password'))
        return super().update(instance, validated_data)


class LoginSerializer(serializers.Serializer):
    username = serializers.CharField()
    password = serializers.CharField()



import os
from unittest import mock

from django.conf import settings


def test_static_files(client):
    root = os.path.join(settings.BASE_DIR, 'utilities/tests/static')
    with mock.patch.object(settings, 'WHITENOISE_ROOT', root):
        # an API request
        response = client.get('/api/')
        assert 'workflows' in response.data

        # an unmatched file, goes to homepage
        response = client.get('/something')
        assert next(response.streaming_content) == b'index\n'

        # a real static file
        response = client.get('/static.txt')
        assert next(response.streaming_content) == b'static\n'



import sys

from unittest import mock

from django.core import management

from yawn.management.commands import webserver


@mock.patch.object(webserver.WSGIApplication, 'run')
def test_webserver(mock_run):
    # pytest with coverage arguments causes gunicorn's argparse to fail
    with mock.patch.object(sys, 'argv', ['yawn', 'worker', '-b', '0.0.0.0:8000']):
        management.call_command('webserver')
        assert mock_run.call_count == 1



from yawn.worker.models import Worker


def test_get_task(client, run):
    task = run.task_set.first()
    worker = Worker.objects.create(name='worker1')
    task.start_execution(worker)
    url = '/api/tasks/%s/' % task.id
    response = client.get(url)
    assert response.status_code == 200, response.data
    assert response.data['name'] == task.template.name
    assert response.data['status'] == task.status
    assert response.data['executions'][0]['worker']['name'] == worker.name
    assert response.data['messages'][0]['queue'] == task.message_set.first().queue.name


def test_patch_task_terminate(client, run):
    task = run.task_set.first()
    url = '/api/tasks/%s/' % task.id
    worker = Worker.objects.create(name='worker1')
    execution = task.start_execution(worker)
    response = client.patch(url, {'terminate': execution.id})
    execution.refresh_from_db()
    assert response.status_code == 200, response.data
    assert response.data['executions'][0]['id'] == execution.id
    assert response.data['executions'][0]['status'] == execution.status
    assert response.data['executions'][0]['status'] == execution.KILLED


def test_patch_task_enqueue(client, run):
    task = run.task_set.first()
    assert task.message_set.count() == 1
    url = '/api/tasks/%s/' % task.id
    response = client.patch(url, {'enqueue': True})
    assert response.status_code == 200, response.data
    assert len(response.data['messages']) == 2


def test_get_execution(client, run):
    task = run.task_set.first()
    worker = Worker.objects.create(name='worker1')
    execution = task.start_execution(worker)
    execution.mark_finished(exit_code=0)
    response = client.get('/api/executions/')
    assert response.status_code == 200, response.data
    executions = response.data['results']
    assert len(executions) == 1
    assert executions[0]['status'] == execution.SUCCEEDED
    assert executions[0]['minutes_running'][:2] == '0m'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-29 16:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('yawn', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='queue',
            name='name',
            field=models.TextField(unique=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='run',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='yawn.Run'),
        ),
        migrations.AlterField(
            model_name='template',
            name='name',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='template',
            name='workflow',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT,
                                    to='yawn.Workflow'),
        ),
        migrations.AlterField(
            model_name='workflowname',
            name='name',
            field=models.TextField(unique=True),
        ),
    ]



import errno
import os
import signal
import subprocess
import select
import time
import typing

from yawn.utilities import logger


class Result:
    """
    Returned to communicate the status of a process

    In practice has stdout/stderr OR returncode, not both
    """
    stdout = None
    stderr = None
    returncode = None

    def __init__(self, execution_id):
        self.execution_id = execution_id

    def __str__(self):
        return 'Result(id={}, stdout={}, stderr={}, returncode={})'.format(
            self.execution_id, self.stdout, self.stderr, self.returncode)


class Execution:
    """Internal process tracker"""

    def __init__(self, process, execution_id, timeout):
        self.id = execution_id
        self.process = process
        self.start_time = time.monotonic()
        self.deadline = self.start_time + timeout if timeout else None

    def __str__(self):
        return 'Result(id={}, stdout={}, stderr={}, returncode={})'.format(
            self.id, self.process.stdout, self.process.stderr, self.process.returncode)


class Manager:
    """
    Class for starting subprocesses and reading from them.
    """

    def __init__(self):
        # dict mapping file descriptors to the owning subprocess
        self.pipes = dict()  # type: typing.Dict[typing.io, Execution]
        self.running = dict()  # type: typing.Dict[int, Execution]

    def start_subprocess(self, execution_id: int, command: str, environment: dict, timeout: int) -> None:
        """
        Start a subprocess:
        - extend the parent process's environment with custom environment variables
        - track stdout and stderr file descriptors for later reading
        - set process group to facilitate killing any children of the command

        :param execution_id: the ID of the Execution instance being run
        :param command: a list of arguments, first argument must be an executable
        :param environment: environment variables from the WorkflowRun
        :param timeout: maximum number of seconds the process should be allowed to run for
        """
        process_environment = os.environ.copy()
        for key, value in environment.items():
            # all variables must be strings, be explicit so it fail in our code
            process_environment[key] = str(value)

        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                   preexec_fn=os.setpgrp, env=process_environment, shell=True)

        # store references to the process and file descriptors
        # Popen gives us io.BufferedReader; get the raw file handle instead
        execution = Execution(process, execution_id, timeout)
        self.pipes.update({
            process.stdout.raw: execution,
            process.stderr.raw: execution
        })
        self.running[execution_id] = execution

    def read_output(self, timeout=0.1) -> typing.List[Result]:
        """
        Read from all ready subprocess file descriptors.

        The returned byte string could include partial utf-8 characters,
        and when decoding the user should trim 0-3 bytes off the end to
        get a readable unicode string.

        :param timeout: in seconds, provided for testing
        """
        all_results = {}
        read_set = self.pipes.keys()

        try:
            # 1/10 second timeout so we return control to the calling event loop
            # if no file descriptors are ready to read
            read_ready, _, _ = select.select(read_set, [], [], timeout)
        except select.error as exc:
            # a signal could interrupt. fixed in python 3.5:
            # https://www.python.org/dev/peps/pep-0475/
            if exc.args[0] != errno.EINTR:
                raise
            logger.warning('Received select.error: %s', exc)
            read_ready = []

        # read from each ready file descriptor
        for fd in read_ready:
            execution = self.pipes[fd]
            data = fd.read(1024)
            if data == b'':
                # the pipe is closed
                fd.close()
                del self.pipes[fd]
                continue

            # keep reading up to 3 more bytes until we get a full UTF-8 character
            for _ in range(3):
                try:
                    data = data.decode('utf-8')
                    break
                except UnicodeDecodeError:
                    data += fd.read1(1)
            else:
                logger.error('Unable to decode byte data! Throwing it away.')

            result = all_results.setdefault(execution.id, Result(execution.id))
            if fd == execution.process.stdout.raw:
                result.stdout = data
            else:
                result.stderr = data

        # check if each running process needs cleanup
        for execution in list(self.running.values()):
            # check if the process should be killed
            if execution.deadline and execution.deadline < time.monotonic():
                # kill the process group, in an attempt to kill any children it has spawned
                try:
                    # setpgrp above sets the PGID to the subprocess' PID
                    os.killpg(execution.process.pid, signal.SIGKILL)
                    logger.info('Terminated execution #%s', execution.id)
                except (ProcessLookupError, PermissionError) as exc:
                    logger.info('Execution #%s was marked to kill but has already exited (%s)',
                                execution.id, exc.__class__.__name__)

            # check if the process has exited
            exit_code = execution.process.poll()
            if exit_code is None:
                continue  # we'll check again later

            # we may not have read everything available, so only cleanup after all pipes are closed
            open_pipes = (
                {execution.process.stdout.raw, execution.process.stderr.raw}
                & set(self.pipes.keys())
            )
            if not open_pipes:
                result = all_results.setdefault(execution.id, Result(execution.id))
                result.returncode = execution.process.returncode
                del self.running[execution.id]

        return list(all_results.values())

    def mark_terminated(self, execution_ids: typing.Iterable[int]):
        """
        Set the deadline to now for the given execution_ids so they
        are terminated in the next loop.
        """
        now = time.monotonic()
        for execution_id in execution_ids:
            if execution_id in self.running:
                self.running[execution_id].deadline = now
            else:
                logger.info('Execution #%s is not running, ignoring termination request', execution_id)

    def get_running_ids(self):
        return list(self.running.keys())



import django
import pytest

from django.conf import settings
from django.db import transaction
from django.test import runner

import yawn.settings.test

# Setup at the module level because django settings need to be
# initialized before importing other django code.
# This file is only imported by pytest when running tests.
settings.configure(**yawn.settings.test.__dict__)
django.setup()


@pytest.fixture(scope='session')
def setup_django():
    """Provide a test database and django configuration"""
    from yawn.worker.models import Queue

    manager = runner.DiscoverRunner(verbosity=1, interactive=False)
    old_config = manager.setup_databases()

    # create the default queue outside the transaction
    Queue.get_default_queue()

    yield

    manager.teardown_databases(old_config)


class TestPassedRollback(Exception):
    """Signal to rollback after a test passes"""


@pytest.fixture(autouse=True)
def test_transaction(setup_django, request):
    if 'no_transaction' in request.keywords:
        # the database reconnect test cannot work inside a transaction
        # so provide a way to disable transactions.
        yield
    else:
        try:
            with transaction.atomic():
                yield
                raise TestPassedRollback()
        except TestPassedRollback:
            pass


@pytest.fixture()
def client():
    from django.contrib.auth.models import User
    from rest_framework.test import APIClient

    user = User.objects.create_user('test_user', is_staff=True)
    api_client = APIClient()
    api_client.force_authenticate(user)
    return api_client


@pytest.fixture()
def run():
    """Setup a workflow and run to test with"""
    from yawn.workflow.models import WorkflowName
    from yawn.task.models import Template

    name = WorkflowName.objects.create(name='workflow1')
    workflow = name.new_version(parameters={'parent': True, 'child': False})
    task1 = Template.objects.create(workflow=workflow, name='task1', command='')
    task2 = Template.objects.create(workflow=workflow, name='task2', command='')
    task2.upstream.add(task1)

    return workflow.submit_run(parameters={'child': True})



import re
from operator import itemgetter

from django.db import transaction
from django.db.models import NOT_PROVIDED
from rest_framework import serializers

from yawn.task.serializers import TaskSerializer
from yawn.worker.models import Queue
from yawn.workflow.models import Workflow, WorkflowName, Run
from yawn.task.models import Template


class WorkflowNameSerializer(serializers.ModelSerializer):
    current_version = serializers.IntegerField(source='current_version.version')
    current_version_id = serializers.IntegerField()
    task_count = serializers.IntegerField()
    schedule = serializers.CharField(source='current_version.schedule')
    schedule_active = serializers.BooleanField(source='current_version.schedule_active')
    versions = serializers.SerializerMethodField()

    class Meta:
        model = WorkflowName
        fields = '__all__'

    def get_versions(self, obj):
        """Provide a full list of versions, only on detail"""
        if self.context['view'].action == 'retrieve':
            return list(obj.workflow_set.order_by('id').values_list('id', flat=True))


class TemplateSerializer(serializers.ModelSerializer):
    queue = serializers.SlugRelatedField(slug_field='name', queryset=Queue.objects.all())
    upstream = serializers.ListField(
        child=serializers.CharField(), default=[], source='upstream.all')

    class Meta:
        model = Template
        exclude = ('workflow',)


class WorkflowSerializer(serializers.ModelSerializer):
    name = serializers.CharField(source='name.name')
    name_id = serializers.IntegerField(read_only=True)
    tasks = TemplateSerializer(many=True, allow_empty=False, source='template_set')

    class Meta:
        model = Workflow
        fields = '__all__'
        read_only_fields = ('next_run', 'name_id')

    @transaction.atomic
    def create(self, validated_data):
        workflow_name = validated_data.pop('name')['name']
        name, _ = WorkflowName.objects.get_or_create(name=workflow_name)
        # get the current version and check for changes
        if name.current_version and unchanged(name.current_version, validated_data):
            # TODO return 302 found instead of 201 created
            return name.current_version

        # changed or no existing version: create a new version
        tasks = validated_data.pop('template_set')
        workflow = name.new_version(**validated_data)
        for task_kwargs in tasks:
            # the awkward 'all' is so serialization gets related tasks through the manager method
            upstream = workflow.template_set.filter(name__in=task_kwargs.pop('upstream')['all'])
            task = Template.objects.create(workflow=workflow, **task_kwargs)
            task.upstream.set(upstream)

        return workflow

    def validate_parameters(self, parameters):
        """Validate variable name is a slug and value is a string"""
        if not isinstance(parameters, dict):
            raise serializers.ValidationError('Workflow parameters must be a dictionary')
        key_regex = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*')
        errors = []
        for key, value in parameters.items():
            if not key_regex.fullmatch(key):
                errors.append('Invalid parameter key: %s' % key)
            if not isinstance(value, str):
                errors.append('Invalid parameter value: %s' % value)
        if errors:
            raise serializers.ValidationError(errors)
        return parameters

    def validate_tasks(self, tasks):
        """Validate tasks only reference already defined upstream tasks"""
        seen_tasks = set()
        errors = []
        for task in tasks:
            name = task['name']
            if name in seen_tasks:
                errors.append('Task names must be unique. Tas %s is defined more than once' % name)
            missing = set(task['upstream']['all']) - seen_tasks
            if missing:
                errors.append(
                    'Task {task} specifies upstream task(s) {upstream} but they do not exist. '
                    'Hint: check the order of your tasks.'.format(
                        task=name, upstream=', '.join(missing))
                )
            seen_tasks.add(name)

        if errors:
            raise serializers.ValidationError(errors)
        return tasks


def unchanged(current_workflow, new_data):
    """
    Detect if the new workflow is unchanged from the current version.

    This is ugly; suggestions would be appreciated...
    """
    ignore = ('id', 'name', 'version', 'schedule_active', 'next_run', 'parameters')
    if not compare_fields(current_workflow, new_data, ignore):
        return False

    new_tasks = sorted(new_data['template_set'], key=itemgetter('name'))
    current_tasks = list(current_workflow.template_set.order_by('name'))
    if len(new_tasks) != len(current_tasks):
        return False

    for current_task, new_task in zip(current_tasks, new_tasks):
        ignore = ('id', 'upstream', 'workflow')
        if not compare_fields(current_task, new_task, ignore):
            return False

        if set(new_task['upstream']['all']) == \
                set(current_task.upstream.values_list('name', flat=True)):
            continue
        else:
            return False

    return True


def compare_fields(instance, new_data, ignore_fields):
    """Compare a model instance to a new dict; return True if they match"""
    for field in instance._meta.fields:
        if field.name in ignore_fields:
            # ignored fields
            continue
        current_value = getattr(instance, field.name)
        if field.name not in new_data:
            # no value provided, check if the current value is the default
            default = field.default if field.default is not NOT_PROVIDED else None
            if current_value == default:
                continue
            else:
                return False
        if current_value != new_data[field.name]:
            return False

    return True


class RunSerializer(serializers.ModelSerializer):
    tasks = TaskSerializer(many=True, source='task_set', read_only=True)
    workflow_id = serializers.IntegerField()

    class Meta:
        model = Run
        exclude = ('workflow',)
        read_only_fields = ('submitted_time', 'scheduled_time', 'status')

    def create(self, validated_data):
        workflow = Workflow.objects.get(id=validated_data['workflow_id'])
        parameters = validated_data.get('parameters')
        run = workflow.submit_run(parameters)
        run.refresh_from_db()  # load submitted_time...
        return run



from rest_framework import serializers

from yawn.worker.models import Worker, Message, Queue


class WorkerSerializer(serializers.ModelSerializer):
    class Meta:
        model = Worker
        fields = '__all__'


class QueueSerializer(serializers.ModelSerializer):
    message_count = serializers.IntegerField(read_only=True, source='message__count')
    purge = serializers.BooleanField(write_only=True, default=False)

    class Meta:
        model = Queue
        fields = '__all__'

    def update(self, instance, validated_data):
        if validated_data['purge'] is True:
            instance.message_set.all().delete()
            instance.message__count = 0
        return instance


class MessageSerializer(serializers.ModelSerializer):
    queue = serializers.CharField(source='queue.name')

    class Meta:
        model = Message
        fields = ('id', 'queue', 'queue_id')



from django.db import models
from django.db import transaction
from django.db.models import functions

from yawn.worker.models import Worker, Queue, Message
from yawn.utilities import logger


class Template(models.Model):
    workflow = models.ForeignKey('Workflow', models.PROTECT, editable=False, null=True)
    queue = models.ForeignKey(Queue, models.PROTECT)
    name = models.TextField()

    command = models.TextField()
    max_retries = models.IntegerField(default=0)
    timeout = models.IntegerField(null=True)  # seconds

    # self-reference for upstream tasks
    upstream = models.ManyToManyField(
        'Template', related_name='downstream', symmetrical=False, blank=True)

    class Meta:
        unique_together = ('workflow', 'name')

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # can't set the default above because the value isn't serializable for migrations
        if not self.queue_id:
            self.queue = Queue.get_default_queue()
        super().save(*args, **kwargs)


class Task(models.Model):
    WAITING = 'waiting'
    QUEUED = 'queued'
    RUNNING = 'running'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    UPSTREAM_FAILED = 'upstream_failed'
    STATUS_CHOICES = [(x, x) for x in (WAITING, QUEUED, RUNNING, SUCCEEDED, FAILED, UPSTREAM_FAILED)]

    run = models.ForeignKey('Run', models.PROTECT, null=True)
    template = models.ForeignKey(Template, models.PROTECT)
    status = models.TextField(choices=STATUS_CHOICES, default=WAITING)

    def __str__(self):
        return '#{task.id} ({task.status})'.format(task=self)

    @classmethod
    def first_queued(cls, queue_ids: list):
        """Get the first task for the given queue names, or None"""
        messages = list(Message.objects.raw(
            # this means retried tasks will go to the top of the queue
            """SELECT * FROM yawn_message WHERE queue_id IN %s
            ORDER BY id LIMIT 1 FOR UPDATE SKIP LOCKED""",
            # psycopg2 handles tuple & list params differently, check the docs:
            params=(tuple(queue_ids),)
        ))
        if not messages:
            return None
        messages[0].delete()
        return messages[0].task

    def enqueue(self):
        """Update the task's status and put a message on the queue"""
        self.status = Task.QUEUED
        self.save()
        Message.objects.create(
            task=self,
            queue=self.template.queue
        )

    def start_execution(self, worker) -> 'Execution':
        self.status = self.RUNNING
        self.save(update_fields=['status'])
        execution = Execution.objects.create(
            task=self,
            worker=worker
        )
        return execution

    def update_downstream(self):
        """
        Enqueue downstream tasks that can run since this one has completed.

        If this runs in a separate transaction from saving the task status,
        then we cannot miss a task because of concurrency.
        """
        downstream = Task.objects.filter(
            run=self.run,
            template__upstream=self.template,
            status__in=[Task.WAITING, Task.UPSTREAM_FAILED]
        ).select_for_update()

        for task in downstream:
            if self.status == Task.SUCCEEDED:
                waiting_for = Task.objects.filter(
                    run=self.run,
                    template__downstream=task.template_id
                ).exclude(status=Task.SUCCEEDED).exists()
                if not waiting_for:
                    logger.info('Submitting %s because all dependencies are complete', task)
                    task.enqueue()

            elif task.status == Task.WAITING:
                # my status is failed or upstream_failed, fail anything waiting downstream
                logger.info('Setting status of %s to upstream_failed', task)
                task.status = Task.UPSTREAM_FAILED
                task.save()
                task.update_downstream()


class Execution(models.Model):
    RUNNING = 'running'
    SUCCEEDED = 'succeeded'
    FAILED = 'failed'
    KILLED = 'killed'
    LOST = 'lost'
    STATUS_CHOICES = [(x, x) for x in (RUNNING, SUCCEEDED, FAILED, KILLED, LOST)]

    task = models.ForeignKey(Task, models.PROTECT)
    worker = models.ForeignKey(Worker, models.PROTECT)
    status = models.TextField(default=RUNNING, choices=STATUS_CHOICES)

    start_timestamp = models.DateTimeField(default=functions.Now)
    stop_timestamp = models.DateTimeField(null=True)

    exit_code = models.IntegerField(null=True)
    stdout = models.TextField(default='', blank=True)
    stderr = models.TextField(default='', blank=True)

    def __str__(self):
        return '#{exec.id} ({exec.status})'.format(exec=self)

    @classmethod
    def update_output(cls, execution_id, stdout, stderr):
        """
        Append to stdout & stderr.
        Use concatenation to efficiently update the fields.
        """
        query = Execution.objects.filter(id=execution_id)
        query.update(
            stdout=functions.Concat('stdout', models.Value(stdout or '')),
            stderr=functions.Concat('stderr', models.Value(stderr or ''))
        )

    def mark_finished(self, exit_code=None, lost=False):
        """
        Update the execution status after it has finished:
         successfully, in error, or because it was lost.

        Also update the task and workflow; re-queue the task if
         it should be retried.
        """
        if lost:
            self.status = Execution.LOST
            self.task.enqueue()

        elif exit_code == 0:
            self.task.status = Task.SUCCEEDED
            self.status = Execution.SUCCEEDED

        else:
            # queue another run if there are remaining retries
            # (current execution is not in count b/c it hasn't been saved yet)
            failed_count = self.task.execution_set.filter(status=Task.FAILED).count()
            if failed_count < self.task.template.max_retries:
                self.task.enqueue()
            else:
                self.task.status = Task.FAILED

            self.status = Execution.FAILED

        if self.task.status != Task.RUNNING:
            self.task.save()
            with transaction.atomic():
                self.task.update_downstream()
            if self.task.run:
                self.task.run.update_status()

        self.stop_timestamp = functions.Now()
        self.exit_code = exit_code
        # need to be careful not to overwrite stdout/stderr
        self.save(update_fields=['status', 'stop_timestamp', 'exit_code'])
